==> tml-0.1.tar.gz: tml-0.1/tml.rb <==
# -*- ruby -*-
# $Id: tml.rb,v 1.6 2003/10/29 14:54:44 tommy Exp $
#
# Copyright (C) 2003 TOMITA Masahiro
# tommy@tmtm.org
#

$var_dir = "/var/spool/tml"
$counter_width = 0
$auto_add_user = true
$smtp_server = "localhost"
$smtp_port = "smtp"
$lock_timeout = 5*60		# 5min

class TML

  class Error < StandardError
    def initialize(id, msg)
      @id, @msg = id, msg
      super msg
    end
    attr_reader :id, :msg
  end

  class Lock
    def initialize(lock_f)
      @lock_fd = File::open(lock_f, File::RDWR|File::CREAT)
      st = Time::now
      while not @lock_fd.flock File::LOCK_EX|File::LOCK_NB do
	sleep 1
	if st+$lock_timeout < Time::now then
	  raise Error::new(:ETIMEDOUT, "lock timed out")
	end
      end
    end

    def unlock()
      @lock_fd.flock File::LOCK_UN
      @lock_fd.close
    end
  end

  class Attr
    def initialize(name, type)
      @name, @type = name, type
      @value = nil
    end

    attr_reader :name, :type, :value

    def set(value)
      case @type
      when :Integer
	if value.kind_of? Integer then
	  @value = value
	elsif value.kind_of? String then
	  unless value =~ /^\d+$/ then
	    raise Error::new(:EINVAL, "#{value}: integer value required")
	  end
	  @value = value.to_i
	else
	  raise Error::new(:EINVAL, "#{value}: invalid type")
	end
      when :Boolean
	if value.kind_of? TrueClass or value.kind_of? FalseClass then
	  @value = value
	elsif value.kind_of? NilClass then
	  @value = false
	elsif value.kind_of? String then
	  case value.downcase
	  when "true", "yes"
	    @value = true
	  when "false", "no"
	    @value = false
	  else
	    raise Error::new(:EINVAL, "#{value}: invalid value")
	  end
	else
	  raise Error::new(:EINVAL, "#{value}: invalid type")
	end
      when :String
	@value = value
      else
	raise "#{value}: invalid type"
      end
    end
  end

  class Passwd

    def initialize()
      dir = "#{$var_dir}/.etc/"
      TML::mkdir_p dir unless File::directory? dir
      @pwd_f = "#{dir}/passwd"
      Lock::new "#{dir}/lock"
      @list = {}
      unless File::exist? @pwd_f then
	File::open(@pwd_f, File::WRONLY|File::CREAT, 0600) do |f|
	  f.write Marshal::dump @list
	end
      end
      @list = Marshal::load File::open(@pwd_f){|f| f.read}
    end

    def have?(addr)
      @list.key? addr
    end

    def list()
      @list.keys.sort
    end

    def add(addr, pw)
      if @list.key? addr then
	raise TML::Error::new(:EEXIST, "#{addr}: already exist")
      end
      @list[addr] = {:passwd => pw}
      flush
    end

    def bye(addr)
      unless @list.key? addr then
	raise TML::Error::new(:ENOENT, "#{addr}: not found")
      end
      @list.delete addr
      flush
    end

    def password(addr, pw)
      unless @list.key? addr then
	raise TML::Error::new(:ENOENT, "#{addr}: not found")
      end
      @list[addr][:passwd] = pw
      flush
    end

    def auth(addr, pw)
      return @list[addr] == pw
    end

    def flush()
      File::open("#{@pwd_f}.new", File::WRONLY|File::CREAT, 0600) do |f|
	f.write Marshal::dump @list
      end
      File::rename(@pwd_f, "#{@pwd_f}.bak")
      File::rename("#{@pwd_f}.new", @pwd_f)
    end

  end

  @@cache = {}
  @@attributes = {
    :sender		=> :String,
    :overview		=> :String,
    :not_member		=> :Boolean,
    :subject_mlname	=> :Boolean,
    :counter_width	=> :Integer,
    :remote_command	=> :Boolean,
    :remote_admin	=> :Boolean,
    :spool		=> :Boolean,
    :web_interface	=> :Integer,
    :member_list	=> :Boolean,
  }

  def TML::mkdir_p(dir)
    dirs = dir.split(/\/+/)
    if dirs[0] == "" then dirs[0] = "/" end
    1.upto(dirs.size) do |i|
      d = dirs[0,i].join("/")
      next if File::exist? d
      Dir::mkdir d
    end
  end

  def TML::list()
    ret = []
    Dir::foreach($var_dir) do |f|
      next if f[0] == ?. or f[0] == ?-
      ret << f
    end
    ret
  end

  def TML::create(mlname, sender=nil, overview=nil)
    if TML::exist? mlname then
      raise TML::Error::new(:EEXIST, "#{mlname}: already exist")
    end

    TML::mkdir_p "#{$var_dir}/#{mlname}"
    File::open("#{$var_dir}/#{mlname}/conf", File::WRONLY|File::CREAT|File::EXCL, 0644).close
    tml = TML::new mlname
    sender = "#{mlname}-admin@#{$domain}" unless sender
    tml.attr[:sender].set sender
    tml.attr[:overview].set(overview || mlname+" ML")
    tml.attr[:not_member].set false
    tml.attr[:subject_mlname].set false
    tml.attr[:counter_width].set $counter_width
    tml.attr[:remote_admin].set false
    tml.attr[:spool].set false
    tml.attr[:web_interface].set 0
    tml.attr[:remote_command].set false
    tml.attr[:member_list].set false
    tml.write_conf
    tml
  end

  def TML::drop(n)
    unless TML::exist? n then
      raise TML::Error::new(:ENOENT, "#{n}: not exist")
    end
    ts = Time::now.strftime("%Y%m%d%H%M%S")
    File::rename("#{$var_dir}/#{n}", "#{$var_dir}/-#{n}-#{ts}")
  end

  def TML::new(n)
    if @@cache.has_key? n then
      return @@cache[n]
    end
    tml = super n
    @@cache[n] = tml
    tml
  end

  def TML::[](n)
    TML::new n
  end

  def TML::exist?(n)
    File::exist? "#{$var_dir}/#{n}/conf"
  end

  def initialize(n)
    unless TML::exist? n then
      err :ENOENT, "#{n}: not exist"
    end
    dir = "#{$var_dir}/#{n}"
    @name = n
    @conf = "#{dir}/conf"
    @cnt_f = "#{dir}/counter"
    @admin_f = "#{dir}/admin"
    @member_f = "#{dir}/member"
    @children_f = "#{dir}/children"
    @parents_f = "#{dir}/parents"
    @counter = nil
    @admins = nil
    @members = nil
    @children = nil
    @parents = nil
    @attr = {}
    @@attributes.each do |n,t|
      @attr[n] = Attr::new(n, t)
    end
    @lock_f = Lock::new "#{dir}/lock"
    read_conf
  end

  def close()
    @lock_f.unlock
    @name = @conf = @cnt_f = @member_f = @children_f = @parents_f = nil
    @counter = @members = @children = @parents = nil
  end

  attr_reader :name, :attr

  def counter()
    return @counter if @counter
    if File::exist? @cnt_f then
      @counter = File::open(@cnt_f, "r") do |f| f.read.to_i end
    else
      File::open(@cnt_f, "w") do |f| f.puts 0 end
      @counter = 0
    end
  end

  def counter=(cnt)
    File::open("#{@cnt_f}.new", "w") do |f| f.puts cnt end
    File::rename("#{@cnt_f}.new", @cnt_f)
    @counter = cnt
  end

  def countup()
    counter
    @counter += 1
    File::open("#{@cnt_f}.new", "w") do |f| f.puts @counter end
    File::rename("#{@cnt_f}.new", @cnt_f)
    @counter
  end

  def members()
    return @members.keys.sort if @members
    @members = {}
    unless File::exist? @member_f then
      return []
    end
    @members = Marshal::load File::open(@member_f){|f| f.read}
    @members.keys.sort
  end

  def all_members()
    m = []
    m += members
    children.each do |c|
      m += c.all_members
    end
    m.uniq
  end

  def have?(addr)
    members.include? addr
  end

  def add(*addrs)
    m = {}
    addrs.flatten.each do |addr|
      if self.have? addr then
	err :EEXIST, "#{addr}: already exist"
      end
      m[addr] = true
    end
    @members.update m
    flush_member
  end

  def bye(*addrs)
    addrs.flatten.each do |addr|
      unless self.have? addr then
	err :ENOENT, "#{addr}: not exist"
      end
    end
    addrs.each do |addr|
      @members.delete addr
    end
    flush_member
  end

  def flush_member()
    File::open("#{@member_f}.new", "w") do |f|
      f.write Marshal::dump @members
    end
    File::rename(@member_f, "#{@member_f}.bak") if File::exist? @member_f
    File::rename("#{@member_f}.new", @member_f)
  end

  def children()
    return @children if @children
    unless File::exist? @children_f then
      @children = []
      return @children
    end
    @children = File::open(@children_f) do |f|
      Marshal::load(f.read).map{|r| TML::new(r)}
    end
  end

  def addml(ml)
    ml = TML::new ml if ml.kind_of? String
    if children.include? ml then
      err :EEXIST, "#{ml.name}: already exist"
    end
    @children << ml
    flush_children
    ml.add_parent self
  end

  def byeml(ml)
    ml = TML::new ml if ml.kind_of? String
    unless children.include? ml then
      err :ENOENT, "#{ml.name}: not exist"
    end
    @children.delete ml
    flush_children
    ml.bye_parent self
  end

  def flush_children()
    File::open("#{@children_f}.new", "w") do |f|
      f.write Marshal::dump @children.map{|c|c.name}
    end
    File::rename(@children_f, "#{@children_f}.bak") if File::exist? @children_f
    File::rename("#{@children_f}.new", @children_f)
  end

  def parents()
    @parents if @parents
    unless File::exist? @parents_f then
      @parents = []
      return @parents
    end
    @parents = File::open(@parents_f) do |f|
      Marshal::load(f.read).map{|r| TML::new(r)}
    end
  end

  def add_parent(ml)
    if parents.include? ml then
      err :EEXIST, "#{ml.name}: already exist"
    end
    @parents << ml
    flush_parents
  end

  def bye_parent(ml)
    unless parents.include? ml then
      err :ENOENT, "#{ml.name}: not exist"
    end
    @parents.delete ml
    flush_parents
  end

  def flush_parents()
    File::open("#{@parents_f}.new", "w") do |f|
      f.write Marshal::dump @parents.map{|c|c.name}
    end
    File::rename(@parents_f, "#{@parents_f}.bak") if File::exist? @parents_f
    File::rename("#{@parents_f}.new", @parents_f)
  end

  def admins()
    return @admins if @admins
    @admins = []
    unless File::exist? @admin_f then
      return @admins
    end
    IO::foreach(@admin_f) do |l|
      next if l[0] == ?#
      @admins << l.chomp
    end
    @admins
  end

  def add_admin(addr)
    if admins.include? addr then
      err :EEXIST, "#{addr}: already exist"
    end
    @admins << addr
    File::open(@admin_f, "a") do |f| f.puts "#{addr}" end
  end

  def bye_admin(addr)
    unless admins.include? addr then
      err :ENOENT, "#{addr}: not exist"
    end
    @admins.delete addr
    File::open("#{@admin_f}.new", "w") do |f|
      IO::foreach(@admin_f) do |line|
	next if line.chomp == addr
	f.puts line
      end
    end
    File::rename(@admin_f, "#{@admin_f}.bak")
    File::rename("#{@admin_f}.new", @admin_f)
  end

  def set_attr(attr, value)
    attr = attr.intern if attr.kind_of? String
    unless @@attributes.key? attr then
      err :EINVAL, "#{attr}: invalid argument"
    end
    @attr[attr].set value
    write_conf
  end

  def get_attr(attr=nil)
    if attr then
      attr = attr.intern if attr.kind_of? String
      unless @@attributes.key? attr then
	err :EINVAL, "#{attr}: unkonwn attribute"
      end
      return @attr[attr].value
    else
      ret = {}
      @attr.each do |n, a|
	ret[n] = a.value
      end
      return ret
    end
  end

  def read_conf()
    IO::foreach(@conf) do |line|
      line = line.chomp.sub(/#.*$/,"")
      next if line =~ /\A\s*\Z/
      var, val = line.strip.split(/\s+/, 2)
      id = var.downcase.intern
      unless @@attributes.key? id then
	err :EINVAL, "unknown parameter: #{var}"
      end
      @attr[id].set val
    end
  end

  def write_conf()
    File::open("#{@conf}.new", File::WRONLY|File::CREAT, 0600) do |f|
      @attr.each do |k,v|
	f.puts "#{k}\t#{v.value}"
      end
    end
    File::rename(@conf, "#{@conf}.bak") if File::exist? @conf
    File::rename("#{@conf}.new", @conf)
  end

  private

  def err(errid, msg)
    raise TML::Error::new(errid, msg)
  end

end

==> tml-0.2.tar.gz: tml-0.2/tml.rb <==
# -*- ruby -*-
# $Id: tml.rb,v 1.8 2003/11/10 06:23:00 tommy Exp $
#
# Copyright (C) 2003 TOMITA Masahiro
# tommy@tmtm.org
#

$var_dir = "/var/spool/tml"
$auto_add_user = true
$smtp_server = "localhost"
$smtp_port = "smtp"
$lock_timeout = 5*60		# 5min
$etc_dir = "#{$var_dir}/.etc"
$auto_aliases = true
$newaliases = nil		# Ex. for postfix "postalias #{$etc_dir}/aliases"

$default_not_member = false
$default_subject_mlname = true
$default_counter_width = 5
$default_remote_command = false
$default_remote_admin = false
$default_spool = true
$default_web_interface = 0
$default_member_list = false

class TML

  class Error < StandardError
    def initialize(id, msg)
      @id, @msg = id, msg
      super msg
    end
    attr_reader :id, :msg
  end

  class Lock
    def initialize(lock_f)
      @lock_fd = File::open(lock_f, File::RDWR|File::CREAT)
      st = Time::now
      while not @lock_fd.flock File::LOCK_EX|File::LOCK_NB do
	sleep 1
	if st+$lock_timeout < Time::now then
	  raise Error::new(:ETIMEDOUT, "lock timed out")
	end
      end
    end

    def unlock()
      @lock_fd.flock File::LOCK_UN
      @lock_fd.close
    end
  end

  class Attr
    def initialize(name, type)
      @name, @type = name, type
      @value = nil
    end

    attr_reader :name, :type, :value

    def set(value)
      case @type
      when :Integer
	if value.kind_of? Integer then
	  @value = value
	elsif value.kind_of? String then
	  unless value =~ /^\d+$/ then
	    raise Error::new(:EINVAL, "#{value}: integer value required")
	  end
	  @value = value.to_i
	else
	  raise Error::new(:EINVAL, "#{value}: invalid type")
	end
      when :Boolean
	if value.kind_of? TrueClass or value.kind_of? FalseClass then
	  @value = value
	elsif value.kind_of? NilClass then
	  @value = false
	elsif value.kind_of? String then
	  case value.downcase
	  when "true", "yes"
	    @value = true
	  when "false", "no"
	    @value = false
	  else
	    raise Error::new(:EINVAL, "#{value}: invalid value")
	  end
	else
	  raise Error::new(:EINVAL, "#{value}: invalid type")
	end
      when :String
	@value = value
      else
	raise "#{value}: invalid type"
      end
    end
  end

  class Passwd

    def initialize()
      TML::mkdir_p $etc_dir unless File::directory? $etc_dir
      @pwd_f = "#{$etc_dir}/passwd"
      Lock::new "#{$etc_dir}/lock"
      @list = {}
      unless File::exist? @pwd_f then
	File::open(@pwd_f, File::WRONLY|File::CREAT, 0600) do |f|
	  f.write Marshal::dump @list
	end
      end
      @list = Marshal::load File::open(@pwd_f){|f| f.read}
    end

    def have?(addr)
      @list.key? addr
    end

    def list()
      @list.keys.sort
    end

    def add(addr, pw)
      if @list.key? addr then
	raise TML::Error::new(:EEXIST, "#{addr}: already exist")
      end
      @list[addr] = {:passwd => pw}
      flush
    end

    def bye(addr)
      unless @list.key? addr then
	raise TML::Error::new(:ENOENT, "#{addr}: not found")
      end
      @list.delete addr
      flush
    end

    def password(addr, pw)
      unless @list.key? addr then
	raise TML::Error::new(:ENOENT, "#{addr}: not found")
      end
      @list[addr][:passwd] = pw
      flush
    end

    def auth(addr, pw)
      return @list[addr] == pw
    end

    def flush()
      File::open("#{@pwd_f}.new", File::WRONLY|File::CREAT, 0600) do |f|
	f.write Marshal::dump @list
      end
      File::rename(@pwd_f, "#{@pwd_f}.bak")
      File::rename("#{@pwd_f}.new", @pwd_f)
    end

  end

  @@cache = {}
  @@attributes = {
    :sender		=> :String,
    :overview		=> :String,
    :not_member		=> :Boolean,
    :subject_mlname	=> :Boolean,
    :counter_width	=> :Integer,
    :remote_command	=> :Boolean,
    :remote_admin	=> :Boolean,
    :spool		=> :Boolean,
    :web_interface	=> :Integer,
    :member_list	=> :Boolean,
  }

  def TML::mkdir_p(dir)
    dirs = dir.split(/\/+/)
    if dirs[0] == "" then dirs[0] = "/" end
    1.upto(dirs.size) do |i|
      d = dirs[0,i].join("/")
      next if File::exist? d
      Dir::mkdir d
    end
  end

  def TML::list()
    ret = []
    Dir::foreach($var_dir) do |f|
      next if f[0] == ?. or f[0] == ?-
      ret << f
    end
    ret
  end

  def TML::create(mlname, admin)
    if TML::exist? mlname then
      raise TML::Error::new(:EEXIST, "#{mlname}: already exist")
    end

    TML::mkdir_p $etc_dir unless File::directory? $etc_dir
    TML::mkdir_p "#{$var_dir}/#{mlname}"
    TML::mkdir_p "#{$var_dir}/#{mlname}/tmp"
    File::open("#{$var_dir}/#{mlname}/conf", File::WRONLY|File::CREAT|File::EXCL, 0644).close
    tml = TML::new mlname
    tml.attr[:sender].set "#{mlname}-admin@#{$domain}"
    tml.attr[:overview].set nil
    tml.attr[:not_member].set $default_not_member
    tml.attr[:subject_mlname].set $default_subject_mlname
    tml.attr[:counter_width].set $default_counter_width
    tml.attr[:remote_admin].set $default_remote_admin
    tml.attr[:spool].set $default_spool
    tml.attr[:web_interface].set $default_web_interface
    tml.attr[:remote_command].set $default_remote_command
    tml.attr[:member_list].set $default_member_list
    tml.write_conf
    tml.add_admin admin
    tml.add_alias if $auto_aliases
    tml
  end

  def TML::drop(n)
    ml = TML::new n
    ml.del_alias if $auto_aliases
    ml.close
    unless TML::exist? n then
      raise TML::Error::new(:ENOENT, "#{n}: not exist")
    end
    ts = Time::now.strftime("%Y%m%d%H%M%S")
    File::rename("#{$var_dir}/#{n}", "#{$var_dir}/-#{n}-#{ts}")
  end

  def TML::new(n)
    if @@cache.has_key? n then
      return @@cache[n]
    end
    tml = super n
    @@cache[n] = tml
    tml
  end

  def TML::[](n)
    TML::new n
  end

  def TML::exist?(n)
    File::exist? "#{$var_dir}/#{n}/conf"
  end

  def initialize(n)
    unless TML::exist? n then
      err :ENOENT, "#{n}: not exist"
    end
    dir = "#{$var_dir}/#{n}"
    @name = n
    @conf = "#{dir}/conf"
    @cnt_f = "#{dir}/counter"
    @admin_f = "#{dir}/admin"
    @member_f = "#{dir}/member"
    @children_f = "#{dir}/children"
    @parents_f = "#{dir}/parents"
    @counter = nil
    @admins = nil
    @members = nil
    @children = nil
    @parents = nil
    @attr = {}
    @@attributes.each do |n,t|
      @attr[n] = Attr::new(n, t)
    end
    @lock_f = Lock::new "#{dir}/lock"
    read_conf

    # clear tmp file
    Dir::glob("#{dir}/tmp/*") do |f|
      File::unlink f if File::mtime(f) < Time::now-24*60*60
    end
  end

  def close()
    @lock_f.unlock
    @name = @conf = @cnt_f = @member_f = @children_f = @parents_f = nil
    @counter = @members = @children = @parents = nil
  end

  attr_reader :name, :attr

  def counter()
    return @counter if @counter
    if File::exist? @cnt_f then
      @counter = File::open(@cnt_f, "r") do |f| f.read.to_i end
    else
      File::open(@cnt_f, "w") do |f| f.puts 0 end
      @counter = 0
    end
  end

  def counter=(cnt)
    File::open("#{@cnt_f}.new", "w") do |f| f.puts cnt end
    File::rename("#{@cnt_f}.new", @cnt_f)
    @counter = cnt
  end

  def countup()
    counter
    @counter += 1
    File::open("#{@cnt_f}.new", "w") do |f| f.puts @counter end
    File::rename("#{@cnt_f}.new", @cnt_f)
    @counter
  end

  def members()
    return @members.keys.sort if @members
    @members = {}
    unless File::exist? @member_f then
      return []
    end
    @members = Marshal::load File::open(@member_f){|f| f.read}
    @members.keys.sort
  end

  def all_members()
    m = []
    m += members
    children.each do |c|
      m += c.all_members
    end
    m.uniq
  end

  def have?(addr)
    members.include? addr
  end

  def add(*addrs)
    m = {}
    addrs.flatten.each do |addr|
      if self.have? addr then
	err :EEXIST, "#{addr}: already exist"
      end
      m[addr] = true
    end
    @members.update m
    flush_member
  end

  def bye(*addrs)
    addrs.flatten.each do |addr|
      unless self.have? addr then
	err :ENOENT, "#{addr}: not exist"
      end
    end
    addrs.each do |addr|
      @members.delete addr
    end
    flush_member
  end

  def flush_member()
    File::open("#{@member_f}.new", "w") do |f|
      f.write Marshal::dump @members
    end
    File::rename(@member_f, "#{@member_f}.bak") if File::exist? @member_f
    File::rename("#{@member_f}.new", @member_f)
  end

  def children()
    return @children if @children
    unless File::exist? @children_f then
      @children = []
      return @children
    end
    @children = File::open(@children_f) do |f|
      Marshal::load(f.read).map{|r| TML::new(r)}
    end
  end

  def addml(ml)
    ml = TML::new ml if ml.kind_of? String
    if children.include? ml then
      err :EEXIST, "#{ml.name}: already exist"
    end
    @children << ml
    flush_children
    ml.add_parent self
  end

  def byeml(ml)
    ml = TML::new ml if ml.kind_of? String
    unless children.include? ml then
      err :ENOENT, "#{ml.name}: not exist"
    end
    @children.delete ml
    flush_children
    ml.bye_parent self
  end

  def flush_children()
    File::open("#{@children_f}.new", "w") do |f|
      f.write Marshal::dump @children.map{|c|c.name}
    end
    File::rename(@children_f, "#{@children_f}.bak") if File::exist? @children_f
    File::rename("#{@children_f}.new", @children_f)
  end

  def parents()
    @parents if @parents
    unless File::exist? @parents_f then
      @parents = []
      return @parents
    end
    @parents = File::open(@parents_f) do |f|
      Marshal::load(f.read).map{|r| TML::new(r)}
    end
  end

  def add_parent(ml)
    if parents.include? ml then
      err :EEXIST, "#{ml.name}: already exist"
    end
    @parents << ml
    flush_parents
  end

  def bye_parent(ml)
    unless parents.include? ml then
      err :ENOENT, "#{ml.name}: not exist"
    end
    @parents.delete ml
    flush_parents
  end

  def flush_parents()
    File::open("#{@parents_f}.new", "w") do |f|
      f.write Marshal::dump @parents.map{|c|c.name}
    end
    File::rename(@parents_f, "#{@parents_f}.bak") if File::exist? @parents_f
    File::rename("#{@parents_f}.new", @parents_f)
  end

  def admins()
    return @admins if @admins
    @admins = []
    unless File::exist? @admin_f then
      return @admins
    end
    IO::foreach(@admin_f) do |l|
      next if l[0] == ?#
      @admins << l.chomp
    end
    @admins
  end

  def add_admin(addr)
    if admins.include? addr then
      err :EEXIST, "#{addr}: already exist"
    end
    @admins << addr
    File::open(@admin_f, "a") do |f| f.puts "#{addr}" end
  end

  def bye_admin(addr)
    unless admins.include? addr then
      err :ENOENT, "#{addr}: not exist"
    end
    if admins.size == 1 then
      err :EBUSY, "one admin is needed at least "
    end
    @admins.delete addr
    File::open("#{@admin_f}.new", "w") do |f|
      IO::foreach(@admin_f) do |line|
	next if line.chomp == addr
	f.puts line
      end
    end
    File::rename(@admin_f, "#{@admin_f}.bak")
    File::rename("#{@admin_f}.new", @admin_f)
  end

  def set_attr(attr, value)
    attr = attr.intern if attr.kind_of? String
    unless @@attributes.key? attr then
      err :EINVAL, "#{attr}: invalid argument"
    end
    @attr[attr].set value
    write_conf
  end

  def get_attr(attr=nil)
    if attr then
      attr = attr.intern if attr.kind_of? String
      unless @@attributes.key? attr then
	err :EINVAL, "#{attr}: unkonwn attribute"
      end
      return @attr[attr].value
    else
      ret = {}
      @attr.each do |n, a|
	ret[n] = a.value
      end
      return ret
    end
  end

  def read_conf()
    IO::foreach(@conf) do |line|
      line = line.chomp.sub(/#.*$/,"")
      next if line =~ /\A\s*\Z/
      var, val = line.strip.split(/\s+/, 2)
      id = var.downcase.intern
      unless @@attributes.key? id then
	err :EINVAL, "unknown parameter: #{var}"
      end
      @attr[id].set val
    end
  end

  def write_conf()
    File::open("#{@conf}.new", File::WRONLY|File::CREAT, 0600) do |f|
      @attr.each do |k,v|
	f.puts "#{k}\t#{v.value}"
      end
    end
    File::rename(@conf, "#{@conf}.bak") if File::exist? @conf
    File::rename("#{@conf}.new", @conf)
  end

  def add_alias()
    lock = Lock::new "#{$etc_dir}/lock"
    file = "#{$etc_dir}/aliases"
    File::open(file, "a") do |f|
      f.puts <<EOS
#{@name}: "|#{$tml_dir}/tml #{@name}"
#{@name}-admin: :include:#{$var_dir}/#{@name}/admin
#{@name}-ctl: "|#{$tml_dir}/tmlctl #{@name}"
EOS
    end
    system $newaliases if $newaliases
    lock.unlock
  end

  def del_alias()
    lock = Lock::new "#{$etc_dir}/lock"
    file = "#{$etc_dir}/aliases"
    File::open(file+".new", "w") do |f|
      IO::foreach(file) do |line|
	next if line =~ /^#{Regexp::escape @name}(-ctl|-admin)?:/
	f.puts line
      end
    end
    File::rename(file+".new", file)
    system $newaliases if $newaliases
    lock.unlock
  end

  def tmpfile(name)
    return "#{$var_dir}/#{@name}/tmp/#{name}"
  end

  def template(name)
    if File::exist? "#{$var_dir}/#{@name}/templates/#{name}" then
      return File::open("#{$var_dir}/#{@name}/templtates/#{name}"){|f| f.read}
    elsif File::exist? "#{$var_dir}/.etc/templates/#{name}" then
      return File::open("#{$var_dir}/.etc/templates/#{name}"){|f| f.read}
    elsif File::exist? "#{$tml_dir}/templates/#{name}" then
      return File::open("#{$tml_dir}/templates/#{name}"){|f| f.read}
    else
      return nil
    end
  end

  private

  def err(errid, msg)
    raise TML::Error::new(errid, msg)
  end

end
